@@ -7,7 +7,7 @@
 from argparse import ArgumentParser, Namespace
 
 from idb.cli import ClientCommand
-from idb.common.types import IdbClient
+from idb.common.types import Client
 
 
 class UrlOpenCommand(ClientCommand):
@@ -23,5 +23,5 @@ class UrlOpenCommand(ClientCommand):
         parser.add_argument("url", help="URL to launch", type=str)
         super().add_parser_arguments(parser)
 
-    async def run_with_client(self, args: Namespace, client: IdbClient) -> None:
+    async def run_with_client(self, args: Namespace, client: Client) -> None:
         await client.open_url(args.url)
