@@ -16,7 +16,7 @@ from io import StringIO
 from pathlib import Path
 from typing import (
     Any,
-    AsyncContextManager,
+    AsyncGenerator,
     AsyncIterable,
     AsyncIterator,
     Dict,
@@ -27,9 +27,11 @@ from typing import (
     Tuple,
 )
 
+import idb.common.plugin as plugin
 from grpclib.client import Channel
 from grpclib.exceptions import GRPCError, ProtocolError, StreamTerminatedError
 from idb.common.constants import TESTS_POLL_INTERVAL
+from idb.common.file import drain_to_file
 from idb.common.gzip import drain_gzip_decompress
 from idb.common.hid import (
     button_press_to_events,
@@ -46,30 +48,37 @@ from idb.common.types import (
     AccessibilityInfo,
     Address,
     AppProcessState,
+    Client as ClientBase,
+    Companion,
     CompanionInfo,
     CrashLog,
     CrashLogInfo,
     CrashLogQuery,
+    DomainSocketAddress,
+    FileContainer,
     FileEntryInfo,
+    FileListing,
     HIDButtonType,
     HIDEvent,
-    IdbClient as IdbClientBase,
     IdbConnectionException,
     IdbException,
     InstalledAppInfo,
     InstalledArtifact,
     InstalledTestInfo,
     InstrumentsTimings,
-    LoggingMetadata,
+    OnlyFilter,
     Permission,
     TargetDescription,
+    TCPAddress,
     TestRunInfo,
+    VideoFormat,
 )
 from idb.grpc.crash import (
     _to_crash_log,
     _to_crash_log_info_list,
     _to_crash_log_query_proto,
 )
+from idb.grpc.file import container_to_grpc as file_container_to_grpc
 from idb.grpc.hid import event_to_grpc
 from idb.grpc.idb_grpc import CompanionServiceStub
 from idb.grpc.idb_pb2 import (
@@ -77,6 +86,7 @@ from idb.grpc.idb_pb2 import (
     AddMediaRequest,
     ApproveRequest,
     ClearKeychainRequest,
+    ConnectRequest,
     ContactsUpdateRequest,
     CrashShowRequest,
     DebugServerRequest,
@@ -89,6 +99,9 @@ from idb.grpc.idb_pb2 import (
     Location,
     LogRequest,
     LsRequest,
+    GetSettingRequest,
+    LOCALE as LocaleSetting,
+    ListSettingRequest,
     MkdirRequest,
     MvRequest,
     OpenUrlRequest,
@@ -100,11 +113,14 @@ from idb.grpc.idb_pb2 import (
     RmRequest,
     ScreenshotRequest,
     SetLocationRequest,
+    SettingRequest,
     TargetDescriptionRequest,
     TerminateRequest,
     UninstallRequest,
+    VideoStreamRequest,
     XctestListBundlesRequest,
     XctestListTestsRequest,
+    XctraceRecordRequest,
 )
 from idb.grpc.install import (
     Bundle,
@@ -125,7 +141,7 @@ from idb.grpc.stream import (
     generate_bytes,
     stop_wrapper,
 )
-from idb.grpc.target import target_to_py
+from idb.grpc.target import companion_to_py, target_to_py
 from idb.grpc.video import generate_video_bytes
 from idb.grpc.xctest import (
     make_request,
@@ -133,24 +149,34 @@ from idb.grpc.xctest import (
     save_attachments,
     write_result_bundle,
 )
+from idb.grpc.xctest_log_parser import XCTestLogParser
+from idb.grpc.xctrace import (
+    xctrace_drain_until_running,
+    xctrace_generate_bytes,
+)
 from idb.utils.contextlib import asynccontextmanager
 
 
-APPROVE_MAP: Dict[Permission, ApproveRequest] = {
+APPROVE_MAP: Dict[Permission, "ApproveRequest.Permission"] = {
     Permission.PHOTOS: ApproveRequest.PHOTOS,
     Permission.CAMERA: ApproveRequest.CAMERA,
     Permission.CONTACTS: ApproveRequest.CONTACTS,
     Permission.URL: ApproveRequest.URL,
     Permission.LOCATION: ApproveRequest.LOCATION,
+    Permission.NOTIFICATION: ApproveRequest.NOTIFICATION,
 }
 
-
-CLIENT_METADATA: LoggingMetadata = {"component": "client"}
+VIDEO_FORMAT_MAP: Dict[VideoFormat, "VideoStreamRequest.Format"] = {
+    VideoFormat.H264: VideoStreamRequest.H264,
+    VideoFormat.RBGA: VideoStreamRequest.RBGA,
+    VideoFormat.MJPEG: VideoStreamRequest.MJPEG,
+    VideoFormat.MINICAP: VideoStreamRequest.MINICAP,
+}
 
 
 def log_and_handle_exceptions(func):  # pyre-ignore
     @functools.wraps(func)
-    @log_call(name=func.__name__, metadata=CLIENT_METADATA)
+    @log_call(name=func.__name__)
     async def func_wrapper(*args: Any, **kwargs: Any) -> Any:  # pyre-ignore
         try:
             return await func(*args, **kwargs)
@@ -162,8 +188,9 @@ def log_and_handle_exceptions(func):  # pyre-ignore
             raise IdbConnectionException(e.strerror)
 
     @functools.wraps(func)
-    @log_call(name=func.__name__, metadata=CLIENT_METADATA)
+    @log_call(name=func.__name__)
     # pyre-fixme[53]: Captured variable `func` is not annotated.
+    # pyre-fixme[3]: Return annotation cannot be `Any`.
     async def func_wrapper_gen(*args: Any, **kwargs: Any) -> Any:
         try:
             async for item in func(*args, **kwargs):
@@ -181,34 +208,98 @@ def log_and_handle_exceptions(func):  # pyre-ignore
         return func_wrapper
 
 
-class IdbClient(IdbClientBase):
+class Client(ClientBase):
     def __init__(
         self,
         stub: CompanionServiceStub,
-        address: Address,
-        is_local: bool,
+        companion: CompanionInfo,
         logger: logging.Logger,
     ) -> None:
         self.stub = stub
-        self.address = address
-        self.is_local = is_local
+        self.companion = companion
         self.logger = logger
 
+    @property
+    def address(self) -> Address:
+        return self.companion.address
+
+    @property
+    def is_local(self) -> bool:
+        return self.companion.is_local
+
     @classmethod
     @asynccontextmanager
     async def build(
-        cls, host: str, port: int, is_local: bool, logger: logging.Logger
-    ) -> AsyncContextManager["IdbClient"]:
-        channel = Channel(host=host, port=port, loop=asyncio.get_event_loop())
-        try:
-            yield IdbClient(
-                stub=CompanionServiceStub(channel=channel),
-                address=Address(host=host, port=port),
-                is_local=is_local,
-                logger=logger,
+        cls,
+        address: Address,
+        logger: logging.Logger,
+        is_local: Optional[bool] = None,
+        exchange_metadata: bool = True,
+        extra_metadata: Optional[Dict[str, str]] = None,
+    ) -> AsyncGenerator["Client", None]:
+        metadata_to_companion = (
+            {
+                **{
+                    key: value
+                    for (key, value) in plugin.resolve_metadata(logger=logger).items()
+                    if isinstance(value, str)
+                },
+                **(extra_metadata or {}),
+            }
+            if exchange_metadata
+            else {}
+        )
+        async with (
+            Channel(host=address.host, port=address.port, loop=asyncio.get_event_loop())
+            if isinstance(address, TCPAddress)
+            else Channel(path=address.path, loop=asyncio.get_event_loop())
+        ) as channel:
+            stub = CompanionServiceStub(channel=channel)
+            with tempfile.NamedTemporaryFile(mode="w+b") as f:
+                try:
+                    response = await stub.connect(
+                        ConnectRequest(
+                            metadata=metadata_to_companion, local_file_path=f.name
+                        )
+                    )
+                except Exception as ex:
+                    raise IdbException(
+                        f"Failed to connect to companion at address {address}: {ex}"
+                    )
+            companion = companion_to_py(
+                companion=response.companion, address=address, is_local=is_local
             )
-        finally:
-            channel.close()
+            if exchange_metadata:
+                metadata_from_companion = {
+                    key: value
+                    for (key, value) in companion.metadata.items()
+                    if isinstance(value, str)
+                }
+                plugin.append_companion_metadata(
+                    logger=logger, metadata=metadata_from_companion
+                )
+            yield Client(stub=stub, companion=companion, logger=logger)
+
+    @classmethod
+    @asynccontextmanager
+    async def for_companion(
+        cls,
+        companion: Companion,
+        udid: str,
+        logger: logging.Logger,
+        only: Optional[OnlyFilter] = None,
+    ) -> AsyncGenerator["Client", None]:
+        with tempfile.NamedTemporaryFile() as temp:
+            # Remove the tempfile so we can bind to it first.
+            os.remove(temp.name)
+            async with companion.unix_domain_server(
+                udid=udid, path=temp.name, only=only
+            ) as resolved_path, Client.build(
+                address=DomainSocketAddress(path=resolved_path),
+                is_local=True,
+                logger=logger,
+            ) as client:
+                yield client
 
     async def _tail_specific_logs(
         self,
@@ -269,8 +360,12 @@ class IdbClient(IdbClientBase):
         self.logger.info(data.strip())
 
     @log_and_handle_exceptions
-    async def list_apps(self) -> List[InstalledAppInfo]:
-        response = await self.stub.list_apps(ListAppsRequest())
+    async def list_apps(
+        self, fetch_process_state: bool = True
+    ) -> List[InstalledAppInfo]:
+        response = await self.stub.list_apps(
+            ListAppsRequest(suppress_process_state=fetch_process_state is False)
+        )
         return [
             InstalledAppInfo(
                 bundle_id=app.bundle_id,
@@ -285,11 +380,18 @@ class IdbClient(IdbClientBase):
 
     @log_and_handle_exceptions
     async def accessibility_info(
-        self, point: Optional[Tuple[int, int]]
+        self, point: Optional[Tuple[int, int]], nested: bool
     ) -> AccessibilityInfo:
         grpc_point = Point(x=point[0], y=point[1]) if point is not None else None
         response = await self.stub.accessibility_info(
-            AccessibilityInfoRequest(point=grpc_point)
+            AccessibilityInfoRequest(
+                point=grpc_point,
+                format=(
+                    AccessibilityInfoRequest.NESTED
+                    if nested
+                    else AccessibilityInfoRequest.LEGACY
+                ),
+            )
         )
         return AccessibilityInfo(json=response.json)
 
@@ -304,7 +406,7 @@ class IdbClient(IdbClientBase):
                 await stream.end()
                 await stream.recv_message()
             else:
-                print(file_paths)
+                self.logger.info(f"Adding media from {file_paths}")
                 generator = stream_map(
                     generate_tar(
                         paths=file_paths,
@@ -358,17 +460,19 @@ class IdbClient(IdbClientBase):
         await self.stub.terminate(TerminateRequest(bundle_id=bundle_id))
 
     @log_and_handle_exceptions
-    async def describe(self) -> TargetDescription:
-        response = await self.stub.describe(TargetDescriptionRequest())
+    async def describe(self, fetch_diagnostics: bool = False) -> TargetDescription:
+        response = await self.stub.describe(
+            TargetDescriptionRequest(fetch_diagnostics=fetch_diagnostics)
+        )
         target = response.target_description
         return target_to_py(
             target=target,
+            # Use the local understanding of the companion instead of the remote's.
             companion=CompanionInfo(
-                host=self.address.host,
-                port=self.address.port,
-                udid=target.udid,
-                is_local=self.is_local,
+                address=self.address, udid=target.udid, is_local=self.is_local
             ),
+            # Extract the companion metadata from the response.
+            metadata=response.companion.metadata,
         )
 
     @log_and_handle_exceptions
@@ -384,23 +488,50 @@ class IdbClient(IdbClientBase):
         await self.stub.uninstall(UninstallRequest(bundle_id=bundle_id))
 
     @log_and_handle_exceptions
-    async def rm(self, bundle_id: str, paths: List[str]) -> None:
-        await self.stub.rm(RmRequest(bundle_id=bundle_id, paths=paths))
-
-    @log_and_handle_exceptions
-    async def mv(self, bundle_id: str, src_paths: List[str], dest_path: str) -> None:
-        await self.stub.mv(
-            MvRequest(bundle_id=bundle_id, src_paths=src_paths, dst_path=dest_path)
+    async def rm(self, container: FileContainer, paths: List[str]) -> None:
+        await self.stub.rm(
+            RmRequest(paths=paths, container=file_container_to_grpc(container))
         )
 
     @log_and_handle_exceptions
-    async def ls(self, bundle_id: str, path: str) -> List[FileEntryInfo]:
-        response = await self.stub.ls(LsRequest(bundle_id=bundle_id, path=path))
+    async def mv(
+        self, container: FileContainer, src_paths: List[str], dest_path: str
+    ) -> None:
+        await self.stub.mv(
+            MvRequest(
+                src_paths=src_paths,
+                dst_path=dest_path,
+                container=file_container_to_grpc(container),
+            )
+        )
+
+    @log_and_handle_exceptions
+    async def ls_single(
+        self, container: FileContainer, path: str
+    ) -> List[FileEntryInfo]:
+        response = await self.stub.ls(
+            LsRequest(path=path, container=file_container_to_grpc(container))
+        )
         return [FileEntryInfo(path=file.path) for file in response.files]
 
     @log_and_handle_exceptions
-    async def mkdir(self, bundle_id: str, path: str) -> None:
-        await self.stub.mkdir(MkdirRequest(bundle_id=bundle_id, path=path))
+    async def ls(self, container: FileContainer, paths: List[str]) -> List[FileListing]:
+        response = await self.stub.ls(
+            LsRequest(paths=paths, container=file_container_to_grpc(container))
+        )
+        return [
+            FileListing(
+                parent=listing.parent.path,
+                entries=[FileEntryInfo(path=entry.path) for entry in listing.files],
+            )
+            for listing in response.listings
+        ]
+
+    @log_and_handle_exceptions
+    async def mkdir(self, container: FileContainer, path: str) -> None:
+        await self.stub.mkdir(
+            MkdirRequest(path=path, container=file_container_to_grpc(container))
+        )
 
     @log_and_handle_exceptions
     async def crash_delete(self, query: CrashLogQuery) -> List[CrashLogInfo]:
@@ -455,11 +586,15 @@ class IdbClient(IdbClientBase):
             yield response
 
     @log_and_handle_exceptions
-    async def push(self, src_paths: List[str], bundle_id: str, dest_path: str) -> None:
+    async def push(
+        self, src_paths: List[str], container: FileContainer, dest_path: str
+    ) -> None:
         async with self.stub.push.open() as stream:
             await stream.send_message(
                 PushRequest(
-                    inner=PushRequest.Inner(bundle_id=bundle_id, dst_path=dest_path)
+                    inner=PushRequest.Inner(
+                        dst_path=dest_path, container=file_container_to_grpc(container)
+                    )
                 )
             )
             if self.is_local:
@@ -480,14 +615,16 @@ class IdbClient(IdbClientBase):
                 )
 
     @log_and_handle_exceptions
-    async def pull(self, bundle_id: str, src_path: str, dest_path: str) -> None:
+    async def pull(
+        self, container: FileContainer, src_path: str, dest_path: str
+    ) -> None:
         async with self.stub.pull.open() as stream:
             request = request = PullRequest(
-                bundle_id=bundle_id,
                 src_path=src_path,
                 # not sending the destination to remote companion
                 # so it streams the file back
                 dst_path=dest_path if self.is_local else None,
+                container=file_container_to_grpc(container),
             )
             await stream.send_message(request)
             await stream.end()
@@ -521,7 +658,7 @@ class IdbClient(IdbClientBase):
         await self.hid(iterator_to_async_iterator(events))
 
     @log_and_handle_exceptions
-    async def tap(self, x: int, y: int, duration: Optional[float] = None) -> None:
+    async def tap(self, x: float, y: float, duration: Optional[float] = None) -> None:
         await self.send_events(tap_to_events(x, y, duration))
 
     @log_and_handle_exceptions
@@ -615,7 +752,7 @@ class IdbClient(IdbClientBase):
         timings: Optional[InstrumentsTimings] = None,
         post_process_arguments: Optional[List[str]] = None,
     ) -> List[str]:
-        self.logger.info(f"Starting instruments connection")
+        self.logger.info("Starting instruments connection")
         async with self.stub.instruments_run.open() as stream:
             self.logger.info("Sending instruments request")
             await stream.send_message(
@@ -689,6 +826,7 @@ class IdbClient(IdbClientBase):
         args: Optional[List[str]] = None,
         env: Optional[Dict[str, str]] = None,
         foreground_if_running: bool = False,
+        wait_for_debugger: bool = False,
         stop: Optional[asyncio.Event] = None,
     ) -> None:
         async with self.stub.launch.open() as stream:
@@ -698,6 +836,7 @@ class IdbClient(IdbClientBase):
                     env=env,
                     app_args=args,
                     foreground_if_running=foreground_if_running,
+                    wait_for_debugger=wait_for_debugger,
                     wait_for=True if stop else False,
                 )
             )
@@ -712,7 +851,7 @@ class IdbClient(IdbClientBase):
 
     @log_and_handle_exceptions
     async def record_video(self, stop: asyncio.Event, output_file: str) -> None:
-        self.logger.info(f"Starting connection to backend")
+        self.logger.info("Starting connection to backend")
         async with self.stub.record.open() as stream:
             if self.is_local:
                 self.logger.info(
@@ -722,7 +861,7 @@ class IdbClient(IdbClientBase):
                     RecordRequest(start=RecordRequest.Start(file_path=output_file))
                 )
             else:
-                self.logger.info(f"Starting video recording with response data")
+                self.logger.info("Starting video recording with response data")
                 await stream.send_message(
                     RecordRequest(start=RecordRequest.Start(file_path=None))
                 )
@@ -740,6 +879,59 @@ class IdbClient(IdbClientBase):
                 )
                 self.logger.info(f"Finished decompression to {output_file}")
 
+    async def stream_video(
+        self,
+        output_file: Optional[str],
+        fps: Optional[int],
+        format: VideoFormat,
+        compression_quality: float,
+        scale_factor: float = 1,
+    ) -> AsyncGenerator[bytes, None]:
+        self.logger.info("Starting connection to backend")
+        async with self.stub.video_stream.open() as stream:
+            if self.is_local and output_file:
+                self.logger.info(
+                    f"Streaming locally with companion writing to {output_file}"
+                )
+                await stream.send_message(
+                    VideoStreamRequest(
+                        start=VideoStreamRequest.Start(
+                            file_path=output_file,
+                            fps=fps,
+                            format=VIDEO_FORMAT_MAP[format],
+                            compression_quality=compression_quality,
+                            scale_factor=scale_factor,
+                        )
+                    )
+                )
+            else:
+                self.logger.info("Starting streaming over the wire")
+                await stream.send_message(
+                    VideoStreamRequest(
+                        start=VideoStreamRequest.Start(
+                            file_path=None,
+                            fps=fps,
+                            format=VIDEO_FORMAT_MAP[format],
+                            compression_quality=compression_quality,
+                            scale_factor=scale_factor,
+                        )
+                    )
+                )
+            try:
+                iterator = generate_bytes(stream=stream, logger=self.logger)
+                if output_file and not self.is_local:
+                    self.logger.info(f"Writing wired bytes to {output_file}")
+                    await drain_to_file(stream=iterator, file_path=output_file)
+                else:
+                    async for data in iterator:
+                        yield data
+            finally:
+                self.logger.info("Stopping video streaming")
+                await stream.send_message(
+                    VideoStreamRequest(stop=VideoStreamRequest.Stop())
+                )
+                await stream.end()
+
     @log_and_handle_exceptions
     async def run_xctest(
         self,
@@ -757,6 +949,7 @@ class IdbClient(IdbClientBase):
         timeout: Optional[int] = None,
         poll_interval_sec: float = TESTS_POLL_INTERVAL,
         report_activities: bool = False,
+        report_attachments: bool = False,
         activities_output_path: Optional[str] = None,
         coverage_output_path: Optional[str] = None,
     ) -> AsyncIterator[TestRunInfo]:
@@ -774,10 +967,14 @@ class IdbClient(IdbClientBase):
                 result_bundle_path=result_bundle_path,
                 timeout=timeout,
                 report_activities=(
-                    report_activities or activities_output_path is not None
+                    report_activities
+                    or activities_output_path is not None
+                    or report_attachments
                 ),
+                report_attachments=report_attachments,
                 collect_coverage=coverage_output_path is not None,
             )
+            log_parser = XCTestLogParser()
             await stream.send_message(request)
             await stream.end()
             async for response in stream:
@@ -786,12 +983,13 @@ class IdbClient(IdbClientBase):
                 for line in [
                     line
                     for lines in response.log_output
-                    # pyre-fixme[10]: Name `lines` is used but not defined.
                     for line in lines.splitlines(keepends=True)
                 ]:
+                    log_parser.parse_streaming_log(line.rstrip())
                     self._log_from_companion(line)
                     if idb_log_buffer:
                         idb_log_buffer.write(line)
+
                 if result_bundle_path:
                     await write_result_bundle(
                         response=response,
@@ -801,7 +999,7 @@ class IdbClient(IdbClientBase):
                 if response.coverage_json and coverage_output_path:
                     with open(coverage_output_path, "w") as f:
                         f.write(response.coverage_json)
-                for result in make_results(response):
+                for result in make_results(response, log_parser):
                     if activities_output_path:
                         save_attachments(
                             run_info=result,
@@ -824,3 +1022,136 @@ class IdbClient(IdbClientBase):
             source=LogRequest.COMPANION, stop=stop, arguments=None
         ):
             yield message
+
+    @log_and_handle_exceptions
+    async def set_hardware_keyboard(self, enabled: bool) -> None:
+        await self.stub.setting(
+            SettingRequest(
+                hardwareKeyboard=SettingRequest.HardwareKeyboard(enabled=enabled)
+            )
+        )
+
+    @log_and_handle_exceptions
+    async def set_locale(self, locale_identifier: str) -> None:
+        await self.stub.setting(
+            SettingRequest(
+                stringSetting=SettingRequest.StringSetting(
+                    setting=LocaleSetting, value=locale_identifier
+                )
+            )
+        )
+
+    @log_and_handle_exceptions
+    async def get_locale(self) -> str:
+        response = await self.stub.get_setting(GetSettingRequest(setting=LocaleSetting))
+        return response.value
+
+    @log_and_handle_exceptions
+    async def list_locale_identifiers(self) -> List[str]:
+        response = await self.stub.list_settings(
+            ListSettingRequest(
+                setting=LocaleSetting,
+            )
+        )
+        return list(response.values)
+
+    @log_and_handle_exceptions
+    async def xctrace_record(
+        self,
+        stop: asyncio.Event,
+        # original 'xctrace record' options
+        output: str,
+        template_name: str,
+        all_processes: bool = False,
+        time_limit: Optional[float] = None,
+        package: Optional[str] = None,
+        process_to_attach: Optional[str] = None,
+        process_to_launch: Optional[str] = None,
+        process_env: Optional[Dict[str, str]] = None,
+        launch_args: Optional[List[str]] = None,
+        target_stdin: Optional[str] = None,
+        target_stdout: Optional[str] = None,
+        # FB options
+        post_args: Optional[List[str]] = None,
+        stop_timeout: Optional[float] = None,
+        # control events
+        started: Optional[asyncio.Event] = None,
+    ) -> List[str]:
+        self.logger.info("Starting xctrace connection")
+        async with self.stub.xctrace_record.open() as stream:
+            self.logger.info("Sending xctrace record request")
+            target = None
+            if all_processes:
+                target = XctraceRecordRequest.Target(all_processes=all_processes)
+            elif process_to_attach:
+                target = XctraceRecordRequest.Target(
+                    process_to_attach=process_to_attach
+                )
+            else:
+                target = XctraceRecordRequest.Target(
+                    launch_process=XctraceRecordRequest.LauchProcess(
+                        process_to_launch=process_to_launch,
+                        launch_args=launch_args,
+                        target_stdin=target_stdin,
+                        target_stdout=target_stdout,
+                        process_env=process_env,
+                    )
+                )
+            await stream.send_message(
+                XctraceRecordRequest(
+                    start=XctraceRecordRequest.Start(
+                        template_name=template_name,
+                        time_limit=time_limit,
+                        package=package,
+                        target=target,
+                    )
+                )
+            )
+            self.logger.info("Starting xctrace record")
+            await xctrace_drain_until_running(stream=stream, logger=self.logger)
+            if started:
+                started.set()
+            self.logger.info("Xctrace record has started, waiting for stop")
+            async for response in stop_wrapper(stream=stream, stop=stop):
+                log = response.log
+                if len(log):
+                    self.logger.info(log.decode())
+            self.logger.info("Stopping xctrace record")
+            await stream.send_message(
+                XctraceRecordRequest(
+                    stop=XctraceRecordRequest.Stop(
+                        timeout=stop_timeout,
+                        args=post_args,
+                    )
+                )
+            )
+            await stream.end()
+
+            result = []
+
+            with tempfile.TemporaryDirectory() as tmp_trace_dir:
+                self.logger.info(f"Writing xctrace data from tar to {tmp_trace_dir}")
+                await drain_untar(
+                    xctrace_generate_bytes(stream=stream, logger=self.logger),
+                    output_path=tmp_trace_dir,
+                )
+                if os.path.exists(
+                    os.path.join(os.path.abspath(tmp_trace_dir), "instrument_data")
+                ):
+                    trace_file = f"{output}.trace"
+                    shutil.copytree(tmp_trace_dir, trace_file)
+                    result.append(trace_file)
+                    self.logger.info(f"Trace written to {trace_file}")
+                else:
+                    # tar is a folder containing one or more trace files
+                    for file in os.listdir(tmp_trace_dir):
+                        _, file_extension = os.path.splitext(file)
+                        tmp_trace_file = os.path.join(
+                            os.path.abspath(tmp_trace_dir), file
+                        )
+                        trace_file = f"{output}{file_extension}"
+                        shutil.move(tmp_trace_file, trace_file)
+                        result.append(trace_file)
+                        self.logger.info(f"Trace written to {trace_file}")
+
+            return result
